@@ -24,6 +24,7 @@
 #include "nv_include.h"
 #include "nv_rop.h"
 #include "nvc0_accel.h"
+#include "nouveau_copy.h"
 
 #define NOUVEAU_BO(a, b, c) (NOUVEAU_BO_##a | NOUVEAU_BO_##b | NOUVEAU_BO_##c)
 
@@ -289,7 +290,7 @@ NVC0EXACopy(PixmapPtr pdpix, int srcX , int srcY,
 
 	BEGIN_NVC0(push, SUBC_2D(NV50_GRAPH_SERIALIZE), 1);
 	PUSH_DATA (push, 0);
-	BEGIN_NVC0(push, SUBC_2D(0x088c), 1);
+	BEGIN_NVC0(push, NV50_2D(BLIT_CONTROL), 1);
 	PUSH_DATA (push, 0);
 	BEGIN_NVC0(push, NV50_2D(BLIT_DST_X), 12);
 	PUSH_DATA (push, dstX);
@@ -573,82 +574,84 @@ NVC0EXAPictTexture(NVPtr pNv, PixmapPtr ppix, PicturePtr ppict, unsigned unit)
 {
 	struct nouveau_bo *bo = nouveau_pixmap_bo(ppix);
 	struct nouveau_pushbuf *push = pNv->pushbuf;
+	uint32_t format;
 
 	/*XXX: Scanout buffer not tiled, someone needs to figure it out */
 	if (!nv50_style_tiled_pixmap(ppix))
 		NOUVEAU_FALLBACK("pixmap is scanout buffer\n");
 
-	PUSH_REFN (push, bo, NOUVEAU_BO_VRAM | NOUVEAU_BO_RD);
-	PUSH_DATAu(push, pNv->scratch, TIC_OFFSET + (unit * 32), 8);
 	switch (ppict->format) {
 	case PICT_a8r8g8b8:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_C3, 8_8_8_8));
+		format = _(B_C0, G_C1, R_C2, A_C3, 8_8_8_8);
 		break;
 	case PICT_a8b8g8r8:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_C3, 8_8_8_8));
+		format = _(R_C0, G_C1, B_C2, A_C3, 8_8_8_8);
 		break;
 	case PICT_x8r8g8b8:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_ONE, 8_8_8_8));
+		format = _(B_C0, G_C1, R_C2, A_ONE, 8_8_8_8);
 		break;
 	case PICT_x8b8g8r8:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_ONE, 8_8_8_8));
+		format = _(R_C0, G_C1, B_C2, A_ONE, 8_8_8_8);
 		break;
 	case PICT_r5g6b5:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_ONE, 5_6_5));
+		format = _(B_C0, G_C1, R_C2, A_ONE, 5_6_5);
 		break;
 	case PICT_a8:
-		PUSH_DATA (push, _(A_C0, B_ZERO, G_ZERO, R_ZERO, 8));
+		format = _(A_C0, B_ZERO, G_ZERO, R_ZERO, 8);
 		break;
 	case PICT_x1r5g5b5:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_ONE, 1_5_5_5));
+		format = _(B_C0, G_C1, R_C2, A_ONE, 1_5_5_5);
 		break;
 	case PICT_x1b5g5r5:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_ONE, 1_5_5_5));
+		format = _(R_C0, G_C1, B_C2, A_ONE, 1_5_5_5);
 		break;
 	case PICT_a1r5g5b5:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_C3, 1_5_5_5));
+		format = _(B_C0, G_C1, R_C2, A_C3, 1_5_5_5);
 		break;
 	case PICT_a1b5g5r5:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_C3, 1_5_5_5));
+		format = _(R_C0, G_C1, B_C2, A_C3, 1_5_5_5);
 		break;
 	case PICT_b5g6r5:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_ONE, 5_6_5));
+		format = _(R_C0, G_C1, B_C2, A_ONE, 5_6_5);
 		break;
 	case PICT_b8g8r8x8:
-		PUSH_DATA (push, _(A_ONE, R_C1, G_C2, B_C3, 8_8_8_8));
+		format = _(A_ONE, R_C1, G_C2, B_C3, 8_8_8_8);
 		break;
 	case PICT_b8g8r8a8:
-		PUSH_DATA (push, _(A_C0, R_C1, G_C2, B_C3, 8_8_8_8));
+		format = _(A_C0, R_C1, G_C2, B_C3, 8_8_8_8);
 		break;
 	case PICT_a2b10g10r10:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_C3, 2_10_10_10));
+		format = _(R_C0, G_C1, B_C2, A_C3, 2_10_10_10);
 		break;
 	case PICT_x2b10g10r10:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_ONE, 2_10_10_10));
+		format = _(R_C0, G_C1, B_C2, A_ONE, 2_10_10_10);
 		break;
 	case PICT_x2r10g10b10:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_ONE, 2_10_10_10));
+		format = _(B_C0, G_C1, R_C2, A_ONE, 2_10_10_10);
 		break;
 	case PICT_a2r10g10b10:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_C3, 2_10_10_10));
+		format = _(B_C0, G_C1, R_C2, A_C3, 2_10_10_10);
 		break;
 	case PICT_x4r4g4b4:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_ONE, 4_4_4_4));
+		format = _(B_C0, G_C1, R_C2, A_ONE, 4_4_4_4);
 		break;
 	case PICT_x4b4g4r4:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_ONE, 4_4_4_4));
+		format = _(R_C0, G_C1, B_C2, A_ONE, 4_4_4_4);
 		break;
 	case PICT_a4r4g4b4:
-		PUSH_DATA (push, _(B_C0, G_C1, R_C2, A_C3, 4_4_4_4));
+		format = _(B_C0, G_C1, R_C2, A_C3, 4_4_4_4);
 		break;
 	case PICT_a4b4g4r4:
-		PUSH_DATA (push, _(R_C0, G_C1, B_C2, A_C3, 4_4_4_4));
+		format = _(R_C0, G_C1, B_C2, A_C3, 4_4_4_4);
 		break;
 	default:
 		NOUVEAU_FALLBACK("invalid picture format, this SHOULD NOT HAPPEN. Expect trouble.\n");
 	}
 #undef _
 
+	PUSH_REFN (push, bo, NOUVEAU_BO_VRAM | NOUVEAU_BO_RD);
+	PUSH_DATAu(push, pNv->scratch, TIC_OFFSET + (unit * 32), 8);
+	PUSH_DATA (push, format);
 	PUSH_DATA (push, bo->offset);
 	PUSH_DATA (push, (bo->offset >> 32) |
 			 (bo->config.nvc0.tile_mode << 18) |
@@ -1023,65 +1026,6 @@ NVC0EXARectM2MF(NVPtr pNv, int w, int h, int cpp,
 	return TRUE;
 }
 
-Bool
-NVC0EXARectCopy(NVPtr pNv, int w, int h, int cpp,
-		struct nouveau_bo *src, uint32_t src_off, int src_dom,
-		int src_pitch, int src_h, int src_x, int src_y,
-		struct nouveau_bo *dst, uint32_t dst_off, int dst_dom,
-		int dst_pitch, int dst_h, int dst_x, int dst_y)
-{
-	struct nouveau_pushbuf *push = pNv->ce_pushbuf;
-	struct nouveau_pushbuf_refn refs[] = {
-		{ src, src_dom | NOUVEAU_BO_RD },
-		{ dst, dst_dom | NOUVEAU_BO_WR },
-	};
-	unsigned exec;
-
-	if (nouveau_pushbuf_space(push, 64, 0, 0) ||
-	    nouveau_pushbuf_refn (push, refs, 2))
-		return FALSE;
-
-	exec = 0x00000000;
-	if (!src->config.nvc0.memtype) {
-		src_off += src_y * src_pitch + src_x * cpp;
-		exec |= 0x00000010;
-	}
-	if (!dst->config.nvc0.memtype) {
-		dst_off += dst_y * dst_pitch + dst_x * cpp;
-		exec |= 0x00000100;
-	}
-
-	BEGIN_NVC0(push, SUBC_COPY(0x0200), 7);
-	PUSH_DATA (push, src->config.nvc0.tile_mode);
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, src_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, src_x * cpp);
-	PUSH_DATA (push, src_y);
-	BEGIN_NVC0(push, SUBC_COPY(0x0220), 7);
-	PUSH_DATA (push, dst->config.nvc0.tile_mode);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, dst_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, dst_x * cpp);
-	PUSH_DATA (push, dst_y);
-	BEGIN_NVC0(push, SUBC_COPY(0x030c), 8);
-	PUSH_DATA (push, (src->offset + src_off) >> 32);
-	PUSH_DATA (push, (src->offset + src_off));
-	PUSH_DATA (push, (dst->offset + dst_off) >> 32);
-	PUSH_DATA (push, (dst->offset + dst_off));
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, w * cpp);
-	PUSH_DATA (push, h);
-	BEGIN_NVC0(push, SUBC_COPY(0x0300), 1);
-	PUSH_DATA (push, exec);
-
-	return TRUE;
-}
-
 Bool
 NVE0EXARectCopy(NVPtr pNv, int w, int h, int cpp,
 		struct nouveau_bo *src, uint32_t src_off, int src_dom,
@@ -1089,52 +1033,8 @@ NVE0EXARectCopy(NVPtr pNv, int w, int h, int cpp,
 		struct nouveau_bo *dst, uint32_t dst_off, int dst_dom,
 		int dst_pitch, int dst_h, int dst_x, int dst_y)
 {
-	struct nouveau_pushbuf *push = pNv->pushbuf;
-	struct nouveau_pushbuf_refn refs[] = {
-		{ src, src_dom | NOUVEAU_BO_RD },
-		{ dst, dst_dom | NOUVEAU_BO_WR },
-	};
-	unsigned exec;
-
-	if (nouveau_pushbuf_space(push, 64, 0, 0) ||
-	    nouveau_pushbuf_refn (push, refs, 2))
-		return FALSE;
-
-	exec = 0x00000206;
-	if (!src->config.nvc0.memtype) {
-		src_off += src_y * src_pitch + src_x * cpp;
-		exec |= 0x00000080;
-	}
-	if (!dst->config.nvc0.memtype) {
-		dst_off += dst_y * dst_pitch + dst_x * cpp;
-		exec |= 0x00000100;
-	}
-
-	BEGIN_NVC0(push, SUBC_COPY(0x0728), 6);
-	PUSH_DATA (push, 0x00001000 | src->config.nvc0.tile_mode);
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, src_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, (src_y << 16) | src_x * cpp);
-	BEGIN_NVC0(push, SUBC_COPY(0x070c), 6);
-	PUSH_DATA (push, 0x000001000 | dst->config.nvc0.tile_mode);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, dst_h);
-	PUSH_DATA (push, 1);
-	PUSH_DATA (push, 0);
-	PUSH_DATA (push, (dst_y << 16) | dst_x * cpp);
-	BEGIN_NVC0(push, SUBC_COPY(0x0400), 8);
-	PUSH_DATA (push, (src->offset + src_off) >> 32);
-	PUSH_DATA (push, (src->offset + src_off));
-	PUSH_DATA (push, (dst->offset + dst_off) >> 32);
-	PUSH_DATA (push, (dst->offset + dst_off));
-	PUSH_DATA (push, src_pitch);
-	PUSH_DATA (push, dst_pitch);
-	PUSH_DATA (push, w * cpp);
-	PUSH_DATA (push, h);
-	BEGIN_NVC0(push, SUBC_COPY(0x0300), 1);
-	PUSH_DATA (push, exec);
-
-	return TRUE;
+	return nouveau_copya0b5_rect(pNv->pushbuf, pNv->NvCOPY, w, h, cpp,
+				     src, src_off, src_dom, src_pitch,
+				     src_h, src_x, src_y, dst, dst_off,
+				     dst_dom, dst_pitch, dst_h, dst_x, dst_y);
 }
